package repast.simphony.query;

import net.sf.cglib.reflect.FastClass;
import net.sf.cglib.reflect.FastMethod;
import org.apache.commons.collections15.Predicate;
import repast.simphony.context.Context;

import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

/**
 * Query that returns any objects in a context with a specified property
 * whose value is less than or equal to some specified value.
 *
 * @author Nick Collier
 * @version $Revision$ $Date$
 */
public class PropertyLessThanEquals<T> extends AbstractPropertyQuery<T> {

	private class LessThanEquals<T> implements Predicate<T> {

		Map<Class, FastMethod> methodMap;

		public LessThanEquals(Map<Class, FastMethod> methodMap) {
			this.methodMap = methodMap;
		}

		public boolean evaluate(T t) {
			FastMethod method = methodMap.get(t.getClass());
			if (method != null) {
				try {
					Object obj = method.invoke(t, AbstractPropertyQuery.EMPTY);
					if (obj == null || propertyValue == null) return false;
					return ((Number) method.invoke(t, AbstractPropertyQuery.EMPTY)).doubleValue() <=
									((Number) propertyValue).doubleValue();
				} catch (InvocationTargetException e) {
					AbstractPropertyQuery.msgCenter.error("Error evaluting property less than or equals predicate", e);
				}
			}
			return false;
		}
	}

	/**
	 * Creates a PropertyLessThan query to query the specified context for objects with
	 * the specified property where that property is less than the specified value.
	 *
	 * @param context
	 * @param propertyName
	 * @param propertyValue
	 * @throws IllegalArgumentException if the specified property is not numeric.
	 */
	public PropertyLessThanEquals(Context<T> context, String propertyName, Number propertyValue) {
		super(context, propertyName, propertyValue);
	}


	/**
	 * Creates a Predicate that tests if a numeric property is less than some value.
	 * @param context
	 * @param propertyName
	 *
	 * @return a Predicate that tests if a numeric property is less than some value.
	 */
	protected Predicate<T> createPredicate(Context<T> context, String propertyName) {
		try {
			Map<Class, FastMethod> methodMap = new HashMap<Class, FastMethod>();
			for (Class clazz : context.getAgentTypes()) {
				PropertyDescriptor[] pds = Introspector.getBeanInfo(clazz, Object.class).getPropertyDescriptors();
				for (PropertyDescriptor pd : pds) {
					if (pd.getName().equalsIgnoreCase(propertyName)) {
						Method readMethod = pd.getReadMethod();
						if (readMethod != null) {
							Class<?> returnType = readMethod.getReturnType();
							if (Number.class.isAssignableFrom(returnType) || primNums.contains(returnType)) {
								methodMap.put(clazz, FastClass.create(readMethod.getDeclaringClass()).getMethod(readMethod));
							} else {
								IllegalArgumentException ex = new IllegalArgumentException("Property '" +
								 propertyName + "' must be a numeric property");
								AbstractPropertyQuery.msgCenter.error("Error while creating Property query", ex);
								throw ex;
							}
						}
						break;
					}
				}
			}

			if (methodMap.size() == 0) propertyNotFound(propertyName);

			return new LessThanEquals<T>(methodMap);
		} catch (IntrospectionException e) {
			e.printStackTrace();
			AbstractPropertyQuery.msgCenter.error("Error while creating Property query", e);
		}
		return null;
	}
}
